#!/usr/bin/env python3

import argparse
import asyncio
import logging
import os.path
import traceback
from importlib.metadata import version

from .ffmpeg import Transcoder


async def run():
    parser = argparse.ArgumentParser("video transcoding")
    parser.add_argument("input_file", help="Input file")
    parser.add_argument("output_dir", help="Outout directory")
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add_argument("-n", "--no_gpu", action="store_true", help="Do not use gpu")
    parser.add_argument("--version", action="version", version=version("videopk"))
    parser.add_argument(
        "-b",
        "--bitrate",
        type=int,
        help="Bitrate, in bps, if not specified, nominal bitrate is calculated (preferred option)",
    )

    args = parser.parse_args()

    transcoder = Transcoder()

    transcoder.parameters.try_gpu = not args.no_gpu

    if args.bitrate is not None:
        transcoder.parameters.auto_bitrate = False
        transcoder.parameters.bitrate = args.bitrate

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    try:
        if not os.path.isdir(args.output_dir):
            os.makedirs(args.output_dir)

        output_file = os.path.join(
            args.output_dir,
            os.path.splitext(os.path.basename(args.input_file))[0] + ".mp4",
        )

        await transcoder.transcode(args.input_file, output_file)

    except Exception as e:
        logging.critical(f"Failed to run: {e}")
        if args.verbose:
            traceback.print_exception(e)


def main():
    asyncio.run(run())



import asyncio

from videopk.ffmpeg import Transcoder


async def main():
    transcoder = Transcoder()

    # Try using the gpu for encoding, will throw an exception if no gpu codec is found
    # Currently, this will throw an FFmpegUnsupportedCodec exception.
    transcoder.parameters.try_gpu = True

    # Will compute the nominal bitrate using average framerate of the file and its resolution
    transcoder.parameters.auto_bitrate = True

    # Output file will be in H.265 (HEVC) format
    await transcoder.transcode("input.mp4", "output.mp4")


if __name__ == "__main__":
    asyncio.run(main())



from enum import Enum
from typing import MutableSequence


class CodecType(Enum):
    AUDIO = 0
    VIDEO = 1
    SUBTITLE = 2
    ATTACHMENT = 3
    DATA = 4


class TranscodingParameters(object):
    """Transcoding parameter"""

    def __init__(self) -> None:
        self.try_gpu = True
        self.auto_bitrate = True
        self.only_video = False
        self.bitrate = 0
        self.auto_bitrate = True


class Codec(object):
    def __init__(self) -> None:
        self.encoders: MutableSequence[str] = []
        self.decoders: MutableSequence[str] = []
        self.__decoding = False
        self.__encoding = False
        self.__type = CodecType.VIDEO
        self.__intra_frame_only = False
        self.__lossy = False
        self.__lossless = False
        self.__name = ""
        self.__description = ""

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str) -> None:
        """
        Sets the name

        :param value: the new value that property name should take
        """
        self.__name = value

    @property
    def description(self) -> str:
        return self.__description

    @description.setter
    def description(self, value: str) -> None:
        self.__description = value

    @property
    def decoding(self) -> bool:
        return self.__decoding

    @decoding.setter
    def decoding(self, value: bool) -> None:
        self.__decoding = value

    @property
    def encoding(self) -> bool:
        return self.__encoding

    @encoding.setter
    def encoding(self, value: bool) -> None:
        self.__encoding = value

    @property
    def type(self) -> CodecType:
        return self.__type

    @type.setter
    def type(self, value: CodecType) -> None:
        self.__type = value

    @property
    def intra_frame_only(self) -> bool:
        return self.__intra_frame_only

    @intra_frame_only.setter
    def intra_frame_only(self, value: bool) -> None:
        self.__intra_frame_only = value

    @property
    def lossy(self) -> bool:
        return self.__lossy

    @lossy.setter
    def lossy(self, value: bool) -> None:
        self.__lossy = value

    @property
    def lossless(self) -> bool:
        return self.__lossless

    @lossless.setter
    def lossless(self, value: bool) -> None:
        self.__lossless = value



import pytest

from videopk.ffmpeg import *
from videopk.types import Codec, CodecType


def test_list_codecs():
    codecs = Codecs.instance().list_codecs()

    assert len(codecs) > 0

    # Find the hevc codec
    hevc = [x for x in codecs if x.name == "hevc"]

    assert len(hevc) == 1


def test_parse_line():
    hevc = Codecs.parse_codec_line(
        " DEV.L. hevc                 H.265 / HEVC (High Efficiency Video Coding) (decoders: hevc hevc_v4l2m2m hevc_cuvid) (encoders: hevc_nvenc hevc_v4l2m2m)"
    )

    assert hevc.encoding
    assert hevc.type == CodecType.VIDEO
    assert hevc.lossy
    assert not hevc.lossless
    assert not hevc.intra_frame_only
    assert hevc.name == "hevc"
    assert (
        hevc.description
        == "H.265 / HEVC (High Efficiency Video Coding) (decoders: hevc hevc_v4l2m2m hevc_cuvid) (encoders: hevc_nvenc hevc_v4l2m2m)"
    )

    hevc = Codecs.parse_codec_line(
        "DEV.L. hevc                 H.265 / HEVC (High Efficiency Video Coding) (decoders: hevc hevc_v4l2m2m hevc_cuvid) (encoders: hevc_nvenc hevc_v4l2m2m)"
    )
    assert hevc.encoding
    assert hevc.type == CodecType.VIDEO
    assert hevc.lossy
    assert not hevc.lossless
    assert not hevc.intra_frame_only
    assert hevc.name == "hevc"
    assert (
        hevc.description
        == "H.265 / HEVC (High Efficiency Video Coding) (decoders: hevc hevc_v4l2m2m hevc_cuvid) (encoders: hevc_nvenc hevc_v4l2m2m)"
    )

    assert len(hevc.encoders) == 2
    assert hevc.encoders == ["hevc_nvenc", "hevc_v4l2m2m"]
    assert len(hevc.decoders) == 3

    hevc = Codecs.parse_codec_line(
        "D.V.L. hevc                 H.265 / HEVC (High Efficiency Video Coding)"
    )

    assert hevc.decoding
    assert not hevc.encoding
    assert hevc.type == CodecType.VIDEO
    assert hevc.lossy
    assert not hevc.lossless
    assert not hevc.intra_frame_only
    assert hevc.name == "hevc"

    assert len(hevc.encoders) == 0
    assert len(hevc.decoders) == 1

    with pytest.raises(TypeError):
        hevc = Codecs.parse_codec_line("test 2")


def test_codec():
    codec = Codec()
    assert codec.encoders == []
    assert codec.decoders == []
    assert not codec.decoding
    assert not codec.encoding
    assert not codec.intra_frame_only
    assert codec.type == CodecType.VIDEO
    assert not codec.lossy
    assert not codec.lossless
    assert codec.name == ""
    assert codec.description == ""

    codec.name = "test"
    assert codec.name == "test"

    codec.description = "un test"
    assert codec.description == "un test"

    codec.type = CodecType.AUDIO
    assert codec.type == CodecType.AUDIO

    codec.type = CodecType.SUBTITLE
    assert codec.type == CodecType.SUBTITLE

    codec.type = CodecType.DATA
    assert codec.type == CodecType.DATA

    codec = Codec()
    codec.decoding = True
    assert codec.decoding

    codec = Codec()
    codec.encoding = True
    assert codec.encoding

    codec.intra_frame_only = True
    assert codec.intra_frame_only

    codec = Codec()
    codec.lossy = True
    assert codec.lossy

    codec = Codec()
    codec.lossless = True
    assert codec.lossless



from asyncio import Future
from typing import Protocol, Sequence

from .types import Codec, TranscodingParameters


class ICodecs(Protocol):
    """Defines a codec repository where the available codecs can be queried."""

    def list_codecs(self) -> Sequence[Codec]:
        """Lists the codecs available on the platform.

        :raises RuntimeError: if an error occurs while querying the codecs.

        :return: the codecs available on the platform
        """
        ...


class ITranscoder(Protocol):
    """Represents a transcoder, used to convert a video."""

    parameters: TranscodingParameters
    """Transcoding parameters
    """

    async def transcode(self, input_file: str, output_file: str) -> None:
        """Transcodes a file using the specified transcoding parameter. This method is meant to be run asynchronously by
        return a Future object.

        :param input_file: the file to transcode
        :param output_file: the target file.

        :raises RuntimeError: if an error occurs (will have more granular exceptions in the future)

        """
        ...



import asyncio
import json
import logging
import os
import re
import subprocess
import tempfile
from asyncio import Future
from typing import Dict, Final, Optional, Sequence

from ffmpeg.asyncio import FFmpeg
from ffmpeg.ffmpeg import types

from .constants import H265_BPP
from .interfaces import ICodecs, ITranscoder
from .types import Codec, CodecType, TranscodingParameters


def ffprobe(input_file: str) -> Dict:
    """Runs ffprobe on a file an returns

    :param input_file: the file to probe.

    :return: a dictionary containing the different properties of the file.
    """
    ffprobe_args = "-v quiet -show_format -show_streams -print_format json".split()

    ffprobe_exec = "ffprobe"

    result = subprocess.run(
        [ffprobe_exec] + ffprobe_args + [input_file], stdout=subprocess.PIPE
    )

    return json.loads(result.stdout)


class Codecs(ICodecs):
    """
    Used to list codecs installed on the platform
    """

    DECODING_POS: Final[int] = 0
    ENCODING_POS: Final[int] = 1
    # Type pos is V, A, S, D
    TYPE_POS: Final[int] = 2
    INTRA_FRAME_POS: Final[int] = 3
    LOSSY_POS: Final[int] = 4
    LOSSLESS_POS: Final[int] = 5

    CAP_FIELD: Final[int] = 1
    NAME_FIELD: Final[int] = 2
    DESC_FIELD: Final[int] = 3

    _instance = None

    def __init__(self):
        raise RuntimeError("Call instance() instead")

    @classmethod
    def instance(cls: type["Codecs"]) -> "Codecs":
        if cls._instance is None:
            cls._instance = cls.__new__(cls)

        return cls._instance

    @staticmethod
    def parse_codec_line(line: str) -> Codec:
        fields = re.search("([^\\s]+)\\s([^\\s]+)\\s*(.+)", line.lstrip().rstrip())

        if fields:
            codec = Codec()

            codec.name = fields[Codecs.NAME_FIELD]
            codec.description = fields[Codecs.DESC_FIELD]

            # Parse the capabilities lines
            if fields[Codecs.CAP_FIELD][Codecs.DECODING_POS] == "D":
                codec.decoding = True

            if fields[Codecs.CAP_FIELD][Codecs.ENCODING_POS] == "E":
                codec.encoding = True

            if fields[Codecs.CAP_FIELD][Codecs.TYPE_POS] == "V":
                codec.type = CodecType.VIDEO
            elif fields[Codecs.CAP_FIELD][Codecs.TYPE_POS] == "A":
                codec.type = CodecType.AUDIO
            elif fields[Codecs.CAP_FIELD][Codecs.TYPE_POS] == "S":
                codec.type = CodecType.SUBTITLE
            elif fields[Codecs.CAP_FIELD][Codecs.TYPE_POS] == "D":
                codec.type = CodecType.DATA
            elif fields[Codecs.CAP_FIELD][Codecs.TYPE_POS] == "T":
                codec.type = CodecType.ATTACHMENT

            if fields[Codecs.CAP_FIELD][Codecs.INTRA_FRAME_POS] == "I":
                codec.intra_frame_only = True

            if fields[Codecs.CAP_FIELD][Codecs.LOSSY_POS] == "L":
                codec.lossy = True

            if fields[Codecs.CAP_FIELD][Codecs.LOSSLESS_POS] == "S":
                codec.lossless = True

            if codec.decoding:
                # Try to match encoders and decoders
                m = re.search(r"decoders:([^\)]*)\)", line.lstrip().rstrip())

                if m is not None:
                    codec.decoders = m.group(1).split()

            if codec.encoding:
                m = re.search(r"encoders:([^\)]*)\)", line.lstrip().rstrip())

                if m is not None:
                    codec.encoders = m.group(1).split()

            if codec.encoding and len(codec.encoders) == 0:
                codec.encoders.append(codec.name)

            if codec.decoding and len(codec.decoders) == 0:
                codec.decoders.append(codec.name)

            return codec
        else:
            raise TypeError("invalid line")

    def list_codecs(self) -> Sequence[Codec]:
        """Lists the codecs installed

        :returns: a list of installed codecs
        :raises CalledProcessError if something goes wrong
        """
        result = subprocess.run(["ffmpeg", "-codecs"], capture_output=True, check=True)

        codecs = []

        for line in result.stdout.decode("utf-8").split("\n"):
            try:
                codecs.append(Codecs.parse_codec_line(line))
            except Exception:
                pass

        return codecs


class Transcoder(ITranscoder):
    """The FFmpeg implementation of interface ITranscoder."""

    def __init__(self) -> None:
        self.parameters = TranscodingParameters()

    async def transcode(self, input_file: str, output_file: str) -> None:
        await asyncio.create_task(self.__do_transcode(input_file, output_file))

    async def __apply_rotation(
        self, input_file: str, output_file: str, info: Dict
    ) -> None:
        if "side_data_list" in info["streams"][0]:
            rotation = info["streams"][0]["side_data_list"][0]["rotation"]

            tmp_output_file = tempfile.NamedTemporaryFile(
                suffix=".mp4", dir=".", delete=False
            ).name

            logging.debug(f"Applying rotation of {rotation} deg")

            ffmpeg_cmd = FFmpeg()

            ffmpeg_cmd.option("y").option("display_rotation", rotation).input(
                input_file
            ).output(tmp_output_file, {"map_metadata": 0, "codec": "copy"})

            await ffmpeg_cmd.execute()

            os.rename(tmp_output_file, output_file)

    async def __do_transcode(self, input_file: str, output_file: str) -> None:
        # Get the info from the file
        info = ffprobe(input_file)
        stream_info = info["streams"][0]
        format_name = info["format"]["format_name"]

        ffmpeg_cmd = FFmpeg()

        if self.parameters.auto_bitrate:
            bitrate = int(
                (
                    stream_info["width"]
                    * stream_info["height"]
                    * eval(stream_info["avg_frame_rate"])
                    * H265_BPP
                )
                / 1000
            )
        else:
            bitrate = self.parameters.bitrate // 1000

        # Default output options
        output_options: dict[str, Optional[types.Option]] = {}
        if self.parameters.try_gpu:
            output_options["c:v"] = "hevc_nvenc"
        else:
            output_options["c:v"] = "libx265"
        #            output_options["c:v"] = "libx264"
        #            output_options["profile:v"] = "high"
        #            output_options["level:v"] = 4.0

        logging.debug(f"format name: {format_name}")

        if "mp4" in format_name:
            output_options["map_metadata"] = 0
            output_options["movflags"] = "use_metadata_tags"

        if self.parameters.only_video:
            output_options["an"] = None
        else:
            output_options["c:a"] = "aac"

        output_options["b:v"] = "{}k".format(bitrate)

        logging.debug(f"output_options: {output_options}")

        ffmpeg_cmd.option("y").option("vsync", 0).input(input_file).output(
            output_file, options=output_options
        )

        if self.parameters.try_gpu:
            ffmpeg_cmd.option("hwaccel", "cuda").option("hwaccel_output_format", "cuda")

        await ffmpeg_cmd.execute()

        if "bit_rate" in stream_info:
            logging.debug(f"original bitrate: {stream_info['bit_rate']}")
        logging.debug(
            f"bitrate for {stream_info['width']}x{stream_info['height']}@{eval(stream_info['r_frame_rate'])}: {bitrate}kbps"
        )

        await self.__apply_rotation(output_file, output_file, info)
